import gcsfs

from .config import is_cloud, get_bucket, require_pipeline


def get_fs(gcs_project="", **kwargs):
    if is_cloud():
        return gcsfs.GCSFileSystem(project=gcs_project, token="cloud", **kwargs)
    else:
        return gcsfs.GCSFileSystem(project=gcs_project, token="google_default", **kwargs)


@require_pipeline("save_to_gcfs")
def save_to_gcfs(
    src_path,
    dst_path,
    gcs_project="cal-itp-data-infra",
    bucket=None,
    use_pipe=False,
    verbose=True,
    **kwargs,
):
    """Convenience function for saving files from disk to google cloud storage.

    Arguments:
        src_path: path to file being saved.
        dst_path: path to bucket subdirectory (e.g. "path/to/dir").
    """

    bucket = get_bucket() if bucket is None else bucket

    full_dst_path = bucket + "/" + str(dst_path)

    fs = get_fs(gcs_project)

    if verbose:
        print("Saving to:", full_dst_path)

    if not use_pipe:
        fs.put(str(src_path), full_dst_path, **kwargs)
    else:
        fs.pipe(str(full_dst_path), src_path, **kwargs)

    return full_dst_path


def read_gcfs(src_path, dst_path=None, gcs_project="", bucket=None, verbose=True):
    """
    Arguments:
        src_path: path to file being read from google cloud.
        dst_path: optional path to save file directly on disk.
    """

    bucket = get_bucket() if bucket is None else bucket

    fs = get_fs(gcs_project)

    full_src_path = bucket + "/" + str(src_path)

    if verbose:
        print(f"Reading file: {full_src_path}")

    if dst_path is None:
        return fs.open(full_src_path)
    else:
        # TODO: in this case, dump directly to disk, rather opening
        raise NotImplementedError()

    return full_src_path



import os
import pandas as pd

from functools import singledispatch
from sqlalchemy.sql.expression import Executable, ClauseElement
from sqlalchemy.ext.compiler import compiles
from sqlalchemy import create_engine, Table, MetaData, sql

from .config import (
    CALITP_BQ_MAX_BYTES,
    CALITP_BQ_LOCATION,
    format_table_name,
    get_project_id,
    require_pipeline,
)


class CreateTableAs(Executable, ClauseElement):
    def __init__(self, name, select, replace=False, if_not_exists=False, partition_by=None):
        self.name = name
        self.select = select
        self.replace = replace
        self.if_not_exists = if_not_exists
        self.partition_by = partition_by


@compiles(CreateTableAs)
def visit_insert_from_select(element, compiler, **kw):
    name = compiler.dialect.identifier_preparer.quote(element.name)
    select = compiler.process(element.select, **kw)
    or_replace = " OR REPLACE" if element.replace else ""
    if_not_exists = " IF NOT EXISTS" if element.if_not_exists else ""

    # TODO: visit partition by clause
    partition_by = f" PARTITION BY {element.partition_by}" if element.partition_by else ""

    return f"""
        CREATE{or_replace} TABLE{if_not_exists} {name}
        {partition_by}
        AS {select}
        """


def get_engine(max_bytes=None):
    max_bytes = CALITP_BQ_MAX_BYTES if max_bytes is None else max_bytes

    cred_path = os.environ.get("CALITP_SERVICE_KEY_PATH")

    # Note that we should be able to add location as a uri parameter, but
    # it is not being picked up, so passing as a separate argument for now.
    return create_engine(
        f"bigquery://{get_project_id()}/?maximum_bytes_billed={max_bytes}",
        location=CALITP_BQ_LOCATION,
        credentials_path=cred_path,
    )


def get_table(table_name, as_df=False):
    engine = get_engine()
    src_table = format_table_name(table_name)

    table = Table(src_table, MetaData(bind=engine), autoload=True)

    if as_df:
        return pd.read_sql_query(table.select(), engine)

    return table


@singledispatch
@require_pipeline("write_table")
def write_table(
    sql_stmt,
    table_name,
    engine=None,
    replace=True,
    simplify=True,
    verbose=False,
    if_not_exists=False,
    partition_by=None,
):
    if engine is None:
        engine = get_engine()

    if not isinstance(sql_stmt, sql.ClauseElement):
        sql_stmt = sql.text(sql_stmt)

    create_tbl = CreateTableAs(
        format_table_name(table_name),
        sql_stmt,
        replace=replace,
        if_not_exists=if_not_exists,
        partition_by=partition_by,
    )

    compiled = create_tbl.compile(
        dialect=engine.dialect,
        # binding literals substitutes in parameters, easier to read
        compile_kwargs={"literal_binds": True},
    )

    if verbose:
        print(compiled)

    return engine.execute(compiled)


@write_table.register(pd.DataFrame)
@require_pipeline("write_table")
def _write_table_df(sql_stmt, table_name, engine=None, replace=True):
    if_exists = "replace" if replace else "fail"
    return sql_stmt.to_gbq(format_table_name(table_name), project_id=get_project_id(), if_exists=if_exists)


def query_sql(fname, write_as=None, replace=False, dry_run=False, as_df=True):
    import yaml
    from jinja2 import Environment
    from .templates import user_defined_filters, user_defined_macros

    if fname.endswith(".yml") or fname.endswith(".yaml"):
        config = yaml.safe_load(open(fname))

        sql_template_raw = config["sql"]
    else:
        sql_template_raw = fname

    env = Environment()
    env.filters = {**env.filters, **user_defined_filters}

    template = env.from_string(sql_template_raw)

    sql_code = template.render({**user_defined_macros})

    if dry_run:
        return sql_code
    elif write_as is not None:
        return write_table(sql_code, write_as, replace=replace)
    else:
        engine = get_engine()

        if as_df:
            return pd.read_sql_query(sql_code, engine)

        return engine.execute(sql_code)


def to_snakecase(df):
    """Convert DataFrame column names to snakecase.

    Note that this function also strips some non-ascii charactures, such as '"'.
    """

    return df.rename(
        columns=lambda s: s.lower()
        .replace(" ", "_")
        .replace("&", "_")
        .replace("(", "_")
        .replace(")", "_")
        .replace(".", "_")
        .replace("-", "_")
        .replace("/", "_")
        .replace('"', "")
        .replace("'", "")
    ).rename(columns=lambda s: "_%s" % s if s[0].isdigit() else s)


def sql_patch_comments(table_name, field_comments, table_comments=None, bq_client=None):
    """Patch an existing table with new column descriptions."""

    if bq_client is None:
        from google.cloud import bigquery

        bq_client = bigquery.Client(project=get_project_id(), location=CALITP_BQ_LOCATION)

    tbl = bq_client.get_table(table_name)
    old_schema = tbl.schema

    # make a copy of old schema, then
    new_schema = []
    for col_entry in old_schema:
        d_entry = col_entry.to_api_repr()
        comment = field_comments.get(d_entry["name"])

        if comment:
            # convert entry to dict, change description field, then recreate
            d_entry["description"] = comment

        # fine to keep as dict, since updating table.schema can handle
        new_schema.append(d_entry)

    tbl.schema = new_schema
    tbl.description = table_comments if table_comments is not None else tbl.description

    bq_client.update_table(tbl, ["schema", "description"])



# flake8: noqa

# +
from ipyleaflet import Map, Marker

center = (52.204793, 360.121558)
m = Map(center=center, zoom=15)
marker = Marker(location=center, draggable=True)
m.add_layer(marker)
display(m)



from calitp.protobuff import get_random_protobuff
import datetime
import typer

app = typer.Typer()


@app.command()
def random_protobuff(
    glob=typer.Argument(
        "*",
        help="A glob matching itp_id/url_number/string.",
    ),
    bucket=typer.Option(
        "gtfs-data",
        help="GCS bucket to search.",
    ),
    date=typer.Option(
        f"{datetime.date.today()}*",
        help="Date glob.",
    ),
    format=typer.Option(
        "protobuff",
        help="format to output, json or protobuff.",
    ),
):
    blob, data, error = get_random_protobuff(
        glob,
        bucket=bucket,
        date=date,
        format=format,
    )
    data = str(data)
    print(f"downloaded {blob}")
    if error:
        print(error)
    lines = data.split("\n")
    if len("\n".join(lines[:20])) > 20 * 80:
        print(data[: 20 * 80])
        print(f"... (only showing {20*80}/{len(data)} characters)")
        return
    print("\n".join(lines[:20]))
    if len(lines) > 20:
        print(f"... (only showing 20/{len(lines)} lines)")


@app.callback()
def callback():
    """
    Pull a random protobuff file from the rt archiver storage.
    without --date defaults to midnignt this morning
    full --date string is like 2022-01-01T10:42:16
    --date must be at least YYYY-MM-DD and will default to zeros after that
    """


app()



#!/usr/bin/env python

import re
import ast
from setuptools import setup, find_namespace_packages

_version_re = re.compile(r"__version__\s+=\s+(.*)")

with open("calitp/__init__.py", "rb") as f:
    version = str(ast.literal_eval(_version_re.search(f.read().decode("utf-8")).group(1)))

setup(
    name="calitp",
    version=version,
    packages=find_namespace_packages(),
    install_requires=[
        "gcsfs",
        "pandas",
        "pandas-gbq",
        "sqlalchemy-bigquery",
        "google-cloud-bigquery",
        "gtfs-realtime-bindings",
    ],
    description="",
    author="",
    author_email="",
    url="https://github.com/cal-itp/calitp-py",
)



import json
from IPython.core.magic import register_cell_magic
from IPython.core.magic_arguments import argument, magic_arguments, parse_argstring
from IPython.display import Markdown, display
from IPython import get_ipython
from calitp.sql import query_sql


@magic_arguments()
@argument(
    "-m",
    "--markdown",
    action="store_true",
    help="Print the code to markdown, in addition to running",
)
@argument("-o", "--output", type=str, help="A variable name to save the result as")
@argument("-q", "--quiet", action="store_true", help="Whether to hide the result printout")
@register_cell_magic
def sql(line, cell):
    # %%sql -m
    # %%sql -o tbl_result
    args = parse_argstring(sql, line)

    if args.markdown:
        # print out a markdown object
        display(Markdown(f"""```SQL\n{cell}\n```"""))

    res = query_sql(cell)

    if args.output:
        # if output variable specified, set it as the result in IPython
        shell = get_ipython()
        shell.user_ns[args.output] = res

    if not args.quiet:
        return res


@register_cell_magic
def capture_parameters(line, cell):
    shell = get_ipython()
    shell.run_cell(cell, silent=True)
    # We assume the last line is a tuple
    tup = [s.strip() for s in cell.strip().split("\n")[-1].split(",")]

    print(json.dumps({identifier: shell.user_ns[identifier] for identifier in tup if identifier}))
